@@ -10,6 +10,26 @@
 
 ;;; Code:
 
+(eval-and-compile
+  (defun undercover-root ()
+    (if load-file-name
+        ;; Cask
+        (file-name-directory
+         (directory-file-name
+          (file-name-directory
+           load-file-name)))
+      ;; Flycheck
+      (file-name-directory
+       (directory-file-name
+        default-directory)))))
+
+(eval-when-compile
+  (add-to-list 'load-path
+               (undercover-root)))
+
+(require 'advice)
+(require 'undercover)
+
 (defconst first-example-library-filename
   "test/first-example-library/first-example-library.el")
 
@@ -23,27 +43,46 @@
 (defmacro with-env-variable (name value &rest body)
   "Set environment variable NAME to VALUE and evaluate BODY."
   (declare (indent 2))
- `(let ((process-environment (undercover--clean-environment)))
-    (setenv ,name ,value)
-    (setq undercover--env nil) ; Clear cached environment
-    (progn ,@body)))
+  `(let ((process-environment (undercover--clean-environment)))
+     (setenv ,name ,value)
+     (setq undercover--env nil) ; Clear cached environment
+     (progn ,@body)))
 
 (with-env-variable "TRAVIS" "true"
   (let ((undercover-force-coverage nil))
     (undercover "test/first-example-library/*.el"
-      (:report-file first-example-library-report-file)
-      (:send-report nil))
-    (ignore-errors (delete-file first-example-library-report-file))
-    (add-to-list 'load-path (file-truename "test/first-example-library"))
-    (require 'first-example-library)))
+                (:report-file first-example-library-report-file)
+                (:send-report nil)
+                (:verbosity 10))
+    (ignore-errors (delete-file first-example-library-report-file))))
 
-(ert-deftest test-001/edebug-handlers-are-setted ()
+(eval-and-compile
+  (add-to-list 'load-path
+               (expand-file-name "test/first-example-library" (undercover-root))))
+(require 'first-example-library)
+
+(ert-deftest test/1-setup/edebug-handlers-are-setted ()
   (if (boundp 'edebug-behavior-alist)
       (should (assq 'undercover edebug-behavior-alist))
     (should (eq 'undercover--stop-point-before (symbol-function 'edebug-before)))
     (should (eq 'undercover--stop-point-after (symbol-function 'edebug-after)))))
 
-(ert-deftest test-002/result-is-correct ()
+(ert-deftest test/1-setup/wildcards ()
+  (should (equal (undercover--wildcards-to-files nil)
+                 nil))
+  (should (equal (undercover--wildcards-to-files '("*.el"))
+                 '("undercover.el")))
+  (should (equal (undercover--wildcards-to-files '("*.el"
+                                                   (:exclude "under*.el")))
+                 nil))
+  (should (equal (undercover--wildcards-to-files '("*.el"
+                                                   (:exclude "under*.el")
+                                                   "underco*.el"))
+                 '("undercover.el")))
+  (should (equal (undercover--wildcards-to-files '((:files "a.el" "b.el" "*?")))
+                 '("a.el" "b.el" "*?"))))
+
+(ert-deftest test/2-run/result-is-correct ()
   (should (= 1.0 (distance '(0 0) '(1 0))))
   (should (= 5.0 (distance '(3 3) '(6 7))))
 
@@ -51,15 +90,15 @@
   (should (= 1 (fib 1)))
   (should (= 8 (fib 6))))
 
-(ert-deftest test-003/functions-are-covered ()
+(ert-deftest test/3-verify/functions-are-covered ()
   (should (get 'distance 'edebug))
   (should (get 'fib 'edebug)))
 
-(ert-deftest test-004/check-distance-stop-points-number-of-covers ()
+(ert-deftest test/3-verify/check-distance-stop-points-number-of-covers ()
   (dolist (stop-point-covers (undercover--stop-points-covers 'distance))
     (should (= stop-point-covers 2))))
 
-(ert-deftest test-005/check-coverage-statistics ()
+(ert-deftest test/3-verify/check-coverage-statistics ()
   (undercover--collect-files-coverage undercover--files)
   (let ((example-library-statistics (gethash (file-truename first-example-library-filename)
                                              undercover--files-coverage-statistics)))
@@ -76,7 +115,7 @@
     (should (= 21 (gethash 26 example-library-statistics)))
     (should (= 12 (gethash 27 example-library-statistics)))))
 
-(ert-deftest test-006/check-environment-variables ()
+(ert-deftest test/3-verify/check-environment-variables ()
   (with-env-variable "TRAVIS" "true"
     (should (eq 'coveralls (undercover--detect-report-format)))))
 
@@ -94,7 +133,7 @@
   (should (= (* multiplier 21) (nth 25 example-library-statistics)))
   (should (= (* multiplier 12) (nth 26 example-library-statistics))))
 
-(ert-deftest test-007/check-coveralls-report ()
+(ert-deftest test/3-verify/check-coveralls-report ()
   (with-env-variable "TRAVIS" "true"
     (ad-deactivate 'undercover-safe-report)
     (undercover-safe-report)
@@ -120,12 +159,12 @@
       (undercover-coveralls--merge-reports report)
       (coveralls--check-lines-statistics 2 (gethash "coverage" file-report)))))
 
-(ert-deftest test-008/should-error ()
+(ert-deftest test/3-verify/should-error ()
   (with-env-variable "TRAVIS" nil
     (should-error (undercover-report))
-    (should-error (undercover--create-coveralls-report))))
+    (should-error (undercover-coveralls--create-report))))
 
-(ert-deftest test-009/check-simplecov-report ()
+(ert-deftest test/3-verify/check-simplecov-report ()
   ;; Don't attempt to merge with report in another format
   (when (file-readable-p first-example-library-report-file)
     (delete-file first-example-library-report-file))
@@ -143,7 +182,7 @@
       (undercover-simplecov--merge-reports reportset)
       (coveralls--check-lines-statistics 2 (gethash file-key coverage)))))
 
-(ert-deftest test-010/check-text-report ()
+(ert-deftest test/3-verify/check-text-report ()
   (let* ((undercover--files-coverage-statistics (make-hash-table :test 'equal))
          (undercover--files (list (file-truename "test/first-example-library/first-example-library.el")))
          (report (undercover-text--create-report)))
@@ -151,7 +190,7 @@
 first-example-library : Percent 100% [Relevant: 10 Covered: 10 Missed: 0]
 "))))
 
-(ert-deftest test-011/check-text-report-file ()
+(ert-deftest test/3-verify/check-text-report-file ()
   (let* ((undercover--files-coverage-statistics (make-hash-table :test 'equal))
          (undercover--files (list (file-truename "test/first-example-library/first-example-library.el")))
          (undercover--report-file-path "/tmp/undercover-text-report.txt"))
@@ -163,4 +202,28 @@ first-example-library : Percent 100% [Relevant: 10 Covered: 10 Missed: 0]
 first-example-library : Percent 100% [Relevant: 10 Covered: 10 Missed: 0]
 ")))))
 
+(ert-deftest test/3-verify/check-lcov-report-file ()
+  (let* ((undercover--files-coverage-statistics (make-hash-table :test 'equal))
+         (undercover--files (list (file-truename "test/first-example-library/first-example-library.el")))
+         (undercover--report-file-path "/tmp/undercover-lcov-report.info"))
+    (when (file-readable-p undercover--report-file-path)
+      (delete-file undercover--report-file-path))
+    (undercover-report 'lcov)
+    (let ((report (with-temp-buffer
+                    (insert-file-contents undercover--report-file-path)
+                    (buffer-substring-no-properties (point-min) (point-max)))))
+      (should (string-equal report (format "SF:%s
+DA:15,2
+DA:16,2
+DA:17,2
+DA:18,2
+DA:19,2
+DA:20,2
+DA:24,27
+DA:25,27
+DA:26,21
+DA:27,12
+end_of_record
+" (car undercover--files)))))))
+
 ;;; first-example-library-test.el ends here
