@@ -43,11 +43,33 @@ If nil, auto-detect from the environment.
 
 Configured using the :report-format configuration option.")
 
+(defvar undercover--report-on-kill t
+  "If non-nil, queue generating a report before Emacs exits.
+
+Configured using the :report-on-kill configuration option.")
+
 (defvar undercover--send-report t
   "If non-nil, upload coverage reports to the configured coverage service.
 
 Configured using the :send-report configuration option.")
 
+(defvar undercover--merge-report t
+  "If non-nil, try to merge coverage information into existing report files.
+
+Configured using the :merge-report configuration option.")
+
+;; Currently used levels:
+;; 1 - non-fatal errors
+;; 2 - warnings
+;; 4 - potentially useful output which differs across Undercover runs
+;; 5 - default
+;; 6 - verbose messages which in some situations may point towards a problem
+;; 7 - verbose informational messages
+(defvar undercover--verbosity 5
+  "Controls the amount of messages produced.
+
+Configured using the :verbosity configuration option.")
+
 (defvar undercover--report-file-path nil
   "The path of the file where the coverage report will be written to.
 
@@ -96,12 +118,19 @@ don't overwrite existing KEYS."
   "Search for and return the list of files matched by WILDCARDS.
 
 Example of WILDCARDS: (\"*.el\" \"subdir/*.el\" (:exclude \"exclude-*.el\"))."
-  (cl-destructuring-bind (exclude-clauses include-wildcards)
-      (--separate (and (consp it) (eq :exclude (car it))) wildcards)
-    (let* ((exclude-wildcards (-mapcat #'cdr exclude-clauses))
-           (exclude-files (-mapcat #'file-expand-wildcards exclude-wildcards))
-           (include-files (-mapcat #'file-expand-wildcards include-wildcards)))
-      (-difference include-files exclude-files))))
+  (let (files)
+    (dolist (wildcard wildcards)
+      (setq files
+            (cond
+             ((stringp wildcard)
+              (-union files (file-expand-wildcards wildcard)))
+             ((and (consp wildcard) (eq :exclude (car wildcard)))
+              (-difference files (-mapcat #'file-expand-wildcards (cdr wildcard))))
+             ((and (consp wildcard) (eq :files (car wildcard)))
+              (-union files (cdr wildcard)))
+             (t
+              (error "UNDERCOVER: Unrecognized wildcard pattern: %S" wildcard)))))
+    files))
 
 (defun undercover--getenv-nonempty (name)
   "Return the value of the environment variable NAME if it exists and is non-empty.
@@ -111,6 +140,12 @@ Otherwise, return nil."
     (when (not (zerop (length value)))
       value)))
 
+(defun undercover--message (level format-string &rest args)
+  "Log a message at the given LEVEL.  FORMAT-STRING and ARGS are as in `message'."
+  (declare (indent 1))
+  (when (<= level undercover--verbosity)
+    (apply #'message (concat "UNDERCOVER: " format-string) args)))
+
 
 ;; ----------------------------------------------------------------------------
 ;; `edebug' related functions and hacks:
@@ -140,7 +175,8 @@ Otherwise, return nil."
 
 (defun undercover--load-file-handler (file)
   "Handle the `load' FILE operation."
-  (let ((edebug-all-defs (undercover--coverage-enabled-p))
+  (undercover--message 7 "Instrumenting %s for collecting coverage information." file)
+  (let ((edebug-all-defs (undercover-enabled-p))
         (load-file-name (file-truename file))
         (load-in-progress t))
     (unwind-protect
@@ -150,17 +186,17 @@ Otherwise, return nil."
       (switch-to-buffer (current-buffer)))))
 
 (defun undercover--show-load-file-error (filename load-error)
-  (message "UNDERCOVER: Error while loading %s for coverage:" filename)
-  (message "UNDERCOVER: %s" (error-message-string load-error))
-  (message "UNDERCOVER: The problem may be due to edebug failing to parse the file.")
-  (message "UNDERCOVER: You can try to narrow down the problem using the following steps:")
-  (message "UNDERCOVER: 1. Open %S in an Emacs buffer;" filename)
-  (message "UNDERCOVER: 2. Run M-: `%s';" "(require 'edebug)")
-  (message "UNDERCOVER: 3. Run M-x `edebug-all-defs';")
-  (message "UNDERCOVER: 4. Run M-x `toggle-debug-on-error'.")
-  (message "UNDERCOVER: 5. Run M-x `eval-buffer';")
-  (message "UNDERCOVER: 6. In the *Backtrace* buffer, find a numeric position,")
-  (message "UNDERCOVER:    then M-x `goto-char' to it."))
+  (undercover--message 1 "Error while loading %s for coverage:" filename)
+  (undercover--message 1 "%s" (error-message-string load-error))
+  (undercover--message 1 "The problem may be due to edebug failing to parse the file.")
+  (undercover--message 1 "You can try to narrow down the problem using the following steps:")
+  (undercover--message 1 "1. Open %S in an Emacs buffer;" filename)
+  (undercover--message 1 "2. Run M-: `%s';" "(require 'edebug)")
+  (undercover--message 1 "3. Run M-x `edebug-all-defs';")
+  (undercover--message 1 "4. Run M-x `toggle-debug-on-error'.")
+  (undercover--message 1 "5. Run M-x `eval-buffer';")
+  (undercover--message 1 "6. In the *Backtrace* buffer, find a numeric position,")
+  (undercover--message 1 "   then M-x `goto-char' to it."))
 
 (defun undercover-file-handler (operation &rest args)
   "Handle the `load' OPERATION.  Ignore all ARGS except first."
@@ -174,6 +210,8 @@ Otherwise, return nil."
 
 (defun undercover--edebug-files (files)
   "Use the `edebug' package to instrument all macros and functions in FILES."
+  (undercover--message 6 "Preparing to instrument %d file%s."
+                       (length files) (if (= (length files) 1) "" "s"))
   (when files
     (let ((regexp (->> (-map #'expand-file-name files) (regexp-opt) (format "^%s$"))))
       (add-to-list 'file-name-handler-alist (cons regexp 'undercover-file-handler)))))
@@ -602,6 +640,8 @@ These values may be overridden through the environment (see
   "Calculate and return a hash table representing Undercover's environment."
   (let ((env (make-hash-table :test 'eq)))
     (undercover--detect-ci env)
+    (undercover--message 7 "Detected CI: %s"
+                         (or (gethash :ci-name env) "None"))
     (undercover--read-env env)
     env))
 
@@ -650,6 +690,8 @@ These values may be overridden through the environment (see
 ;; ----------------------------------------------------------------------------
 ;; Coverage format / service support:
 
+;; coveralls.io report:
+
 (defun undercover-coveralls--configured-p ()
   "Check if we can submit a report to Coveralls with what we have/know."
   (cl-case (gethash :ci-type (undercover--need-env))
@@ -682,8 +724,6 @@ These values may be overridden through the environment (see
           (push remote-table remotes-info))))))
 
 
-;; coveralls.io report:
-
 (defun undercover-coveralls--update-report-with-env (report env)
   "Update test coverage REPORT for coveralls.io with information from ENV."
   (undercover--add-to-hash-table report
@@ -824,21 +864,27 @@ These values may be overridden through the environment (see
 
 (defun undercover-coveralls--merge-reports (report)
   "Merge test coverage REPORT with existing from `undercover--report-file-path'."
-  (ignore-errors
-    (let* ((json-object-type 'hash-table)
-           (json-array-type 'list)
-           (old-report (json-read-file undercover--report-file-path))
-           (new-source-files-report (gethash "source_files" report)))
-      (dolist (old-file-hash (gethash "source_files" old-report))
-        (undercover-coveralls--merge-report-file-coverage
-         old-file-hash new-source-files-report)))))
+  (condition-case merge-error
+      (let* ((json-object-type 'hash-table)
+             (json-array-type 'list)
+             (old-report (json-read-file undercover--report-file-path))
+             (new-source-files-report (gethash "source_files" report)))
+        (undercover--message 7
+          "Merging existing Coveralls report: %s" undercover--report-file-path)
+        (dolist (old-file-hash (gethash "source_files" old-report))
+          (undercover-coveralls--merge-report-file-coverage
+           old-file-hash new-source-files-report)))
+    (error
+     (undercover--message 6
+       "Failed to merge Coveralls report: %s" merge-error))))
 
 (defun undercover-coveralls--create-report ()
   "Create test coverage report for coveralls.io."
   (undercover--collect-files-coverage undercover--files)
   (let ((report (make-hash-table :test 'equal)))
     (undercover-coveralls--fill-report report)
-    (undercover-coveralls--merge-reports report)
+    (when undercover--merge-report
+      (undercover-coveralls--merge-reports report))
     (json-encode report)))
 
 (defun undercover-coveralls--save-report (json-report)
@@ -852,16 +898,22 @@ These values may be overridden through the environment (see
   (let* ((coveralls-endpoint (or (getenv "COVERALLS_ENDPOINT")
                                  "https://coveralls.io"))
          (coveralls-url (concat coveralls-endpoint "/api/v1/jobs")))
-    (message "Sending: report to coveralls.io")
+    (undercover--message 5 "Uploading report to coveralls.io")
     (unless (zerop (call-process "curl"
-                                 nil '(:file "/dev/stderr") t
+                                 nil
+                                 (if (>= undercover--verbosity 4)
+                                     '(:file "/dev/stderr")
+                                   nil)
+                                 t
                                  ;; "-v" "--include"
                                  "--fail" "--silent" "--show-error"
                                  "--form"
                                  (concat "json_file=@" undercover--report-file-path)
                                  coveralls-url))
-      (error "Upload to coveralls.io failed"))
-    (message "\nSending: OK")))
+      (error "UNDERCOVER: Upload to coveralls.io failed"))
+    ;; curl's output doesn't end with a newline; print one to stderr now
+    (external-debugging-output ?\n)
+    (undercover--message 5 "Upload OK")))
 
 (defun undercover-coveralls--report ()
   "Create and submit test coverage report to coveralls.io."
@@ -884,6 +936,63 @@ These values may be overridden through the environment (see
 Please disable the :send-report option and use CodeCov's upload
 script (https://codecov.io/bash) instead"))))
 
+;; LCOV report:
+
+(defconst undercover-lcov--test-name nil
+  "The name of the test in the LCOV result set report (for the \"TN:\" line).
+
+If set to nil (the default), no \"TN:\" line will be generated.")
+
+(defun undercover-lcov--create-report ()
+  "Create LCOV test coverage report."
+  (when (and undercover--merge-report
+             (file-readable-p undercover--report-file-path))
+    (user-error "Merging of LCOV reports is not implemented. Please delete %s or invoke with (:merge-report nil)."
+                undercover--report-file-path))
+  (undercover--collect-files-coverage undercover--files)
+  (apply #'concat
+         ;; Test name
+         (if undercover-lcov--test-name
+             (concat "TN:" undercover-lcov--test-name "\n")
+           "")
+         ;; One section per file
+         (mapcar
+          (lambda (file)
+            (let ((statistics (gethash file undercover--files-coverage-statistics))
+                  line-numbers)
+              ;; Collect line numbers
+              (maphash (lambda (k _v)
+                         (push k line-numbers))
+                       statistics)
+              ;; Emit coverage
+              (concat
+               ;; File name
+               "SF:" file "\n"
+               ;; Per-line coverage
+               (apply #'concat
+                      (mapcar
+                       (lambda (line)
+                         (format "DA:%d,%d\n"
+                                 line
+                                 (gethash line statistics)))
+                       (sort line-numbers #'<)))
+               "end_of_record\n")))
+          undercover--files)))
+
+(defun undercover-lcov--report ()
+  "Create test coverage report in LCOV format."
+  (when undercover--send-report
+    (error "UNDERCOVER: Cannot upload LCOV reports.
+
+Please disable the :send-report option (or specify a coverage
+provider as the :report-format instead of 'lcov)."))
+  (let ((undercover--report-file-path (or undercover--report-file-path
+                                          "coverage/lcov.info")))
+    (make-directory (or (file-name-directory undercover--report-file-path) "") t)
+    (with-temp-buffer
+      (insert (undercover-lcov--create-report))
+      (write-region nil nil undercover--report-file-path))))
+
 ;; SimpleCov report:
 
 (defconst undercover-simplecov--report-name "undercover.el"
@@ -928,6 +1037,8 @@ script (https://codecov.io/bash) instead"))))
 (defun undercover-simplecov--merge-reports (new-report)
   "Merge test coverage NEW-REPORT with existing from `undercover--report-file-path'."
   (when (file-readable-p undercover--report-file-path)
+    (undercover--message 7
+      "Merging existing SimpleCov report: %s" undercover--report-file-path)
     (let* ((json-object-type 'hash-table)
            (json-array-type 'list)
            (old-report (json-read-file undercover--report-file-path))
@@ -946,7 +1057,8 @@ script (https://codecov.io/bash) instead"))))
   (undercover--collect-files-coverage undercover--files)
   (let ((report (make-hash-table :test 'equal)))
     (undercover-simplecov--fill-report report)
-    (undercover-simplecov--merge-reports report)
+    (when undercover--merge-report
+      (undercover-simplecov--merge-reports report))
     (json-encode report)))
 
 (defun undercover-simplecov--save-report (json-report)
@@ -1006,10 +1118,11 @@ script (https://codecov.io/bash) instead"))))
 
 (defun undercover--report ()
   "Generate and save/upload a test coverage report, as configured."
-
+  (undercover--message 7 "Generating report.")
   (cl-case undercover--report-format
     ((nil)     (error "UNDERCOVER: Report format not configured and auto-detection failed"))
     (coveralls (undercover-coveralls--report))
+    (lcov      (undercover-lcov--report))
     (simplecov (undercover-simplecov--report))
     (codecov   (undercover-codecov--report))
     (text      (undercover-text--report))
@@ -1034,8 +1147,8 @@ To do this, add `undercover-safe-report' to `kill-emacs-hook'."
 ;; ----------------------------------------------------------------------------
 ;;; Main functions:
 
-(defun undercover--coverage-enabled-p ()
-  "Check if `undercover' is enabled."
+(defun undercover-enabled-p ()
+  "Return non-nil if `undercover' is enabled."
   (or undercover-force-coverage (undercover--under-ci-p)))
 
 (defun undercover-report (&optional report-format)
@@ -1049,8 +1162,8 @@ configuration."
                                            undercover--report-format
                                            (undercover--detect-report-format))))
         (undercover--report))
-    (message
-     "UNDERCOVER: No coverage information. Make sure that your files are not compiled?")))
+    (undercover--message 1
+      "No coverage information. Make sure that your files are not compiled?")))
 
 (defun undercover--env-configuration ()
   "Read configuration from UNDERCOVER_CONFIG."
@@ -1059,39 +1172,46 @@ configuration."
       (condition-case nil
           (car (read-from-string configuration))
         (error
-         (error "UNDERCOVER: error while parsing configuration"))))))
+         (error "UNDERCOVER: Error while parsing configuration"))))))
 
 (defun undercover--set-options (configuration)
   "Extract options from CONFIGURATION and set global variables accordingly.
 
 Options are filtered out, leaving only wildcards, which are returned."
   (cl-destructuring-bind (wildcards options)
-      (--separate (or (stringp it) (eq :exclude (car-safe it))) configuration)
+      (--separate (or (stringp it)
+                      (eq :exclude (car-safe it))
+                      (eq :files (car-safe it)))
+                  configuration)
     (cl-dolist (option options wildcards)
       (cl-case (car-safe option)
+        (:verbosity (setq undercover--verbosity (cadr option)))
         (:report-file (setq undercover--report-file-path (cadr option)))
         (:report-format (setq undercover--report-format (cadr option)))
+        (:report-on-kill (setq undercover--report-on-kill (cadr option)))
         (:send-report (setq undercover--send-report (cadr option)))
+        (:merge-report (setq undercover--merge-report (cadr option)))
         ;; Note: this option is obsolete and intentionally undocumented.
-        ;; Please use :report-file and :send-report explicitly instead.
-        (:report-type (message "UNDERCOVER: The :report-type option is deprecated.")
+        ;; Please use (:report-format 'codecov) (:send-report nil) instead.
+        (:report-type (undercover--message 2 "The :report-type option is deprecated.")
                       (cl-case (cadr option)
                         (:coveralls (setq undercover--report-format 'coveralls))
                         (:codecov (setq undercover--report-format 'codecov)
                                   (setq undercover--send-report nil))
-                        (otherwise (error "Unsupported report-type: %s" (cadr option)))))
-        (otherwise (error "Unsupported option: %s" option))))))
+                        (otherwise (error "UNDERCOVER: Unsupported report-type: %s" (cadr option)))))
+        (otherwise (error "UNDERCOVER: Unsupported option: %s" option))))))
 
 (defun undercover--setup (configuration)
   "Enable test coverage for files matched by CONFIGURATION."
-  (when (undercover--coverage-enabled-p)
+  (when (undercover-enabled-p)
     (let ((env-configuration (undercover--env-configuration))
           (default-configuration '("*.el")))
       (undercover--set-edebug-handlers)
-      (undercover-report-on-kill)
       (let ((wildcards (undercover--set-options
                         (or (append configuration env-configuration)
                             default-configuration))))
+	(when undercover--report-on-kill
+          (undercover-report-on-kill))
         (undercover--edebug-files (undercover--wildcards-to-files wildcards))))))
 
 ;;;###autoload
@@ -1112,20 +1232,44 @@ Undercover is enabled if any of the following is true:
 Each item of CONFIGURATION can be one of the following:
 
 STRING                  Indicates a wildcard of Emacs Lisp files
-                        to include in the coverage.  Examples:
-                        \"*.el\" \"subdir/*.el\"
+                        to include in the coverage.  These are
+                        globbed using `file-expand-wildcards'.
+                        Examples: \"*.el\" \"subdir/*.el\"
 
 (:exclude STRING)       Indicates a wildcard of Emacs Lisp files
-                        to exclude form the coverage.
+                        to exclude from the coverage.
                         Example: (:exclude \"exclude-*.el\")
 
+(:files STRING...)      Indicates a list of Emacs Lisp files to
+                        include in the coverage.  These are
+                        interpreted verbatim and are not globbed.
+
+(:verbosity NUMBER)     Controls how detailed undercover.el
+                        should be in reporting what it's doing
+                        using messages, as a number from 0 (no
+                        messages, fatal errors only) to 10 (all
+                        messages).  The default is 5.
+
 (:report-file STRING)   Sets the path of the file where the
                         coverage report will be written to.
 
+(:report-on-kill BOOLEXP)  Sets whether to queue generating and
+                        saving/uploading a repot before Emacs
+                        exits.  Enabled by default.
+
+                        If disabled, this can still be done by
+                        calling `undercover-report-on-kill'.
+
 (:send-report BOOLEXP)  Sets whether to upload the report to the
                         detected/configured coverage service
                         after generating it.  Enabled by default.
 
+(:merge-report BOOLEXP) If possible, merge collected coverage
+                        data into any existing coverage report
+                        file.  Enabled by default.  If disabled,
+                        undercover.el will always overwrite files
+                        when saving reports.
+
 (:report-format SYMBOL) Sets the report target (file format or
                         coverage service), i.e., what to do with
                         the collected coverage information.
@@ -1144,7 +1288,7 @@ nil          Detect an appropriate service automatically.
              with the CodeCov upload script
              (https://codecov.io/bash).
 
-             Because CodeCov natively understands SimpleCov's
+             Because CodeCov natively understands Coveralls'
              report format, all this does (compared to
              'coveralls) is configure the default save path to a
              location that the upload script will look for.
@@ -1152,6 +1296,9 @@ nil          Detect an appropriate service automatically.
              Uploading from within Undercover is currently not
              supported, and will raise an error.
 
+'lcov        Save the coverage information in the format used by
+             GCOV / LCOV / geninfo.
+
 'simplecov   Save the coverage information as a SimpleCov
              .resultset.json file.
 
@@ -1168,7 +1315,11 @@ Options may also be specified via the environment variable
 \"UNDERCOVER_CONFIG\", which should be formatted as a literal
 Emacs Lisp list consisting of items as defined above.
 Configuration options in \"UNDERCOVER_CONFIG\" override those in
-CONFIGURATION."
+CONFIGURATION.
+
+If no CONFIGURATION is specified (either as an argument, or via
+the environment variable), the default configuration '(\"*.el\")
+is used."
   `(undercover--setup
     (list
      ,@(--map (if (atom it) it `(list ,@it))
